#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import argparse
import json
import logging
import os
import sys
import time

from cupstream2distro import (launchpadmanager, packageinppamanager)
from cupstream2distro.packageinppa import PackageInPPA
from cupstream2distro.packagemanager import list_packages_info_in_str
from cupstream2distro.settings import (
    TIME_BETWEEN_PPA_CHECKS, TIME_BEFORE_STOP_LOOKING_FOR_SOURCE_PUBLISH)

sys.path.append(os.path.join(os.path.dirname(__file__), '../ci-utils'))
from ci_utils import dump_stack


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="Watch for published package in a ppa",
        epilog="""series and ppa options can be set by the corresponding long
        option name env variables as well""")
    parser.add_argument("-u", "--upload-list",
                        help="""A json formated list of packages to check.
                        The format is [{"name": package_name,
                        "version": package_version, "architecture": arch}]""")
    parser.add_argument("-s", "--series",
                        help="Serie used to build the package")
    parser.add_argument("-p", "--ppa",
                        help="""PPA to publish this package to (for instance:
                        'ubuntu-unity/daily-build')""")
    parser.add_argument("-a", "--arch",
                        default=None,
                        help="Only consider the provided target")
    parser.add_argument("-d", "--destppa",
                        help="""Consider this destppa instead of
                        {series}-proposed""")
    return parser.parse_args()


def get_environment_variables(args):
    series = args.series
    ppa = args.ppa
    if not series:
        series = os.getenv("series")
    if not ppa:
        ppa = os.getenv("ppa")
    return (series, ppa)


def get_ppa(ppa):
    return launchpadmanager.get_ppa(ppa)


def watch_ppa(time_start, series, ppa, dest_ppa, arch, upload_list):
    # Prepare launchpad connection:
    lp_series = launchpadmanager.get_series(series)
    monitored_ppa = launchpadmanager.get_ppa(ppa)
    if dest_ppa:
        dest_archive = get_ppa(dest_ppa)
    else:
        dest_archive = launchpadmanager.get_ubuntu_archive()
    logging.info('Series: {}'.format(lp_series))
    logging.info('Monitoring PPA: {}'.format(monitored_ppa))

    logging.info('Destination Archive: {}'.format(dest_archive))

    # Get archs available and archs to ignore
    (available_archs_in_ppa,
     arch_all_arch) = launchpadmanager.get_available_and_all_archs(
        lp_series, monitored_ppa)
    (archs_to_eventually_ignore,
     archs_to_unconditionally_ignore) = launchpadmanager.get_ignored_archs()
    logging.info('Arches available in ppa: {}'.format(available_archs_in_ppa))
    logging.info('All arch in ppa: {}'.format(arch_all_arch))
    logging.info('Arches to eventually ignore: {}'.format(
        archs_to_eventually_ignore))
    logging.info('Arches to unconditionally ignore: {}'.format(
        archs_to_unconditionally_ignore))

    # Collecting all packages that have been uploaded to the ppa
    packages_not_in_ppa = set()
    packages_building = set()
    packages_failed = set()
    for source_package in upload_list:
        source = source_package['name']
        version = source_package['version']
        archs = source_package['architecture']
        logging.info('Inspecting upload: {} - {}'.format(source, version))
        packages_not_in_ppa.add(PackageInPPA(source, version, monitored_ppa,
                                             dest_archive, lp_series,
                                             available_archs_in_ppa,
                                             arch_all_arch,
                                             archs_to_eventually_ignore,
                                             archs_to_unconditionally_ignore,
                                             package_archs=archs))

    # packages_not_in_ppa are packages that were uploaded and are expeceted
    # to eventually appear in the ppa.
    logging.info('Packages not in PPA: {}'.format(
        list_packages_info_in_str(packages_not_in_ppa)))
    logging.info('Packages building: {}'.format(packages_building))
    logging.info('Packages failed: {}'.format(packages_failed))

    # Check the status regularly on all packages
    # TODO The following is the original check loop. This can be extracted
    #    and optimized.
    logging.info("Checking the status for {}".format(
        list_packages_info_in_str(
            packages_not_in_ppa.union(packages_building))))
    packageinppamanager.update_all_packages_status(
        packages_not_in_ppa, packages_building, packages_failed, arch)

    status = {'pending': packages_not_in_ppa,
              'building': packages_building,
              'failed': packages_failed}
    # if we have no package building or failing and have wait for
    # long enough to have some package appearing in the ppa, exit
    if (packages_not_in_ppa and not packages_building and
        ((time.time() - time_start) >
         TIME_BEFORE_STOP_LOOKING_FOR_SOURCE_PUBLISH)):
        # TODO return error on the missing packages
        logging.info(
            "Some source packages were never published in the ppa: "
            "{}".format(list_packages_info_in_str(packages_not_in_ppa)))
        return (1, status)

    # break out of status check loop if all packages have arrived in
    # the ppa and have completed building
    if not packages_not_in_ppa and not packages_building:
        if packages_failed:
            # TODO Return package failure info
            logging.info(
                "Some of the packages failed to build: {}".format(
                    list_packages_info_in_str(packages_failed)))
            return (1, status)
        return (0, status)

    # -1 indicates to retry
    # TODO return useful status about what is still in progress
    return (-1, status)


def main():
    '''Provides usage through the command line.'''
    logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s "
                        "%(message)s")

    args = parse_arguments()
    (series, ppa) = get_environment_variables(args)
    # upload_list is a json formated text field which contains:
    #   [{"name": source_package_name,
    #     "version": source_package_version,
    #     "architecture": architecture_to_monitor}]
    # Example:
    #   [{"name": "cupstream2distro-config",
    #     "version": "0.3.1",
    #     "architecture": "all"}]
    upload_list = json.loads(args.upload_list)

    if not series or not ppa:
        logging.error("Missing compulsory environment variables (ppa, series) "
                      "watching: {}, series: {}".format(ppa, series))
        return 1

    time_start = time.time()
    while True:
        (ret, status) = watch_ppa(time_start, series, ppa,
                                  args.destppa, args.arch, upload_list)
        if ret == -1:
            time.sleep(TIME_BETWEEN_PPA_CHECKS)
        else:
            return ret


if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    sys.exit(main())
