#!/usr/bin/python

import argparse
import os
import sys
import subprocess
import stat
import atexit
import shutil
import StringIO
import tarfile
import tempfile
import textwrap
from distutils import spawn
from itertools import chain

# the juju-deployer/ directory.
deployer_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
# ci-utils probably isn't in our pythonpath
sys.path.append(os.path.join(os.path.dirname(__file__), '../ci-utils'))
from ci_utils import dump_stack


def check_environment():
    '''Cheetah requires environment variables for populating our templates'''
    cheetah_vars = {
        'OS_USERNAME': None,
        'CI_CODE_SOURCE': 'branch',
        'CI_BRANCH': 'lp:ubuntu-ci-services-itself',
        'CI_PAYLOAD_URL': 'n/a',
        'CI_LAUNCHPAD_USER': None,
        'CI_OAUTH_CONSUMER_KEY': None,
        'CI_OAUTH_TOKEN': None,
        'CI_OAUTH_TOKEN_SECRET': None,
        'CI_PPA': 'ppa:canonical-ci-engineering/ci-airline-phase-0',
    }
    missing_required = []
    for key in cheetah_vars:
        if key not in os.environ or len(os.environ[key]) == 0:
            default = cheetah_vars[key]
            if default:
                os.environ[key] = default
            else:
                missing_required.append(key)

    if len(missing_required) > 0:
        template = textwrap.dedent('''
        ERROR: Missing required environment variables:
          {}
        Please ensure the novarc file has been sourced.''')
        exit(template.format('\n  '.join(missing_required)))


def check_bootstrapped():
    try:
        # XXX: Added a print to give user feedback in case sshuttle isn't
        # running and it "hangs", maybe correct behavior is to add a timeout
        # here, but too much work for the benefit, so I'm only providing a
        # hint now -- Ursinha (2014-02-17)
        print 'Checking juju status'
        subprocess.check_output(['juju', 'status'])
    except subprocess.CalledProcessError:
        # There's no need to print an error message as juju status already
        # does that.
        sys.exit(1)


def check_dependencies():
    '''Exit the application if we are missing any dependency programs'''

    needed_commands = ['cheetah']
    for cmd in needed_commands:
        if not spawn.find_executable(cmd):
            err = "%s is missing. Please install it before continuing." % cmd
            print >>sys.stderr, err
            sys.exit(1)


def populate_templates(deployer_dir, working_dir):
    # The command to populate the template.
    cmd_base = ['cheetah', 'fill', '--env', '-p']

    for tmpl in os.listdir(deployer_dir):
        path = os.path.join(deployer_dir, tmpl)
        if os.path.isdir(path):
            src = os.path.join(deployer_dir, tmpl)
            dst = os.path.join(working_dir, tmpl)
            populate_templates(src, dst)
            continue
        if not tmpl.endswith('.yaml.tmpl'):
            continue
        # Strip the .tmpl off .yaml.tmpl and write a novarc-populated version
        # of the yaml to our working directory.
        with open(os.path.join(working_dir, tmpl[:-5]), 'w') as out:
            args = [os.path.join(deployer_dir, tmpl)]
            subprocess.check_call(cmd_base + args, stdout=out)


def generate_argument_list(d, specific=None):
    for f in os.listdir(d):
        if not f.endswith('.yaml'):
            continue
        # If we're looking for a specific set of components to deploy and this
        # isn't one of them, continue.
        if specific and f[:-5] not in specific:
            continue
        yield ('-c', os.path.join(d, f))


def append_python_path(path):
    path = os.path.abspath(path)
    if os.environ.get('PYTHONPATH'):
        os.environ['PYTHONPATH'] += ':%s' % path
    else:
        os.environ['PYTHONPATH'] = '%s' % path


def install(branch, revno=0):
    """Put a copy of a bzr branch in a temporary directory."""

    temp_dir = tempfile.mkdtemp()
    try:
        cmd = ['bzr', 'co', '--lightweight']
        if revno > 0:
            cmd.append('-r%d' % revno)
        cmd += [branch, temp_dir]
        subprocess.check_call(cmd)
        append_python_path(temp_dir)
    except:
        shutil.rmtree(temp_dir)
        raise

    return temp_dir


def install_deployer():
    """Install juju-deployer from bzr. We need juju-deployer from tip to have
    JUJU_REPOSITORY work properly (r87)."""

    deployer = install('lp:juju-deployer', revno=103)
    os.environ['PATH'] = '%s:%s' % (deployer, os.environ['PATH'])
    deployer_bin = os.path.join(deployer, 'juju-deployer')
    with open(deployer_bin, 'w') as fp:
        tmpl = '#!/bin/sh\nPYTHONPATH=%s\nexec python %s/deployer/cli.py $*'
        fp.write(tmpl % (os.environ['PYTHONPATH'], deployer))
    st = os.stat(deployer_bin)
    mode = st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    os.chmod(deployer_bin, mode)
    return deployer


def remove_branch(branch_dir):
    """Clean up the bzr branch if we didn't already do so as a result of
    bzr failing."""

    if os.path.exists(branch_dir):
        shutil.rmtree(branch_dir)


def build_charms(charmsdir=None):
    # juju needs to know about our local charms.
    # and some of our charms require us to call "make" on them
    if not charmsdir:
        charmsdir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), '../charms'))
    os.environ['JUJU_REPOSITORY'] = charmsdir

    def _cleanup(local_charms):
        # delete any charms pulled down by juju-deployer into our local tree
        if not os.path.exists(charmsdir):
            return  # handle special case for unit test harness
        for release in local_charms:
            charms = os.listdir(os.path.join(charmsdir, release))
            remove = set(charms) - set(local_charms[release])
            for r in remove:
                shutil.rmtree(os.path.join(charmsdir, release, r))

    # Feed local_charms with our charms so that if anything goes wrong during
    # the building phase, we'll never remove them.
    local_charms = {'precise': ['lander-jenkins', 'python-django',
                                'rabbitmq-worker', 'restish', 'webui']}
    clean_func = lambda: _cleanup(local_charms)
    atexit.register(clean_func)

    for release in os.listdir(charmsdir):
        if local_charms.get(release) is None:
            local_charms[release] = []
        for charm in os.listdir(os.path.join(charmsdir, release)):
            local_charms[release].append(charm)
            charmdir = os.path.join(charmsdir, release, charm)
            if os.path.exists(os.path.join(charmdir, 'Makefile')):
                print('Building charm: %s' % charm)
                subprocess.check_call(['make'], cwd=charmdir)
    return clean_func


def setup_swift():
    base = os.path.join(os.path.dirname(__file__), '..')
    print('Uploading local branch to swift ...')

    ciutils = os.path.join(base, 'ci-utils')
    sys.path.append(ciutils)
    from ci_utils import data_store

    config = {
        'auth_url': os.environ['OS_AUTH_URL'],
        'auth_user': os.environ['OS_USERNAME'],
        'auth_password': os.environ['OS_PASSWORD'],
        'auth_tenant_name': os.environ['OS_TENANT_NAME'],
        'auth_region': os.environ['OS_REGION_NAME']
    }
    ds = data_store.DataStore('ci-payload', config, public=True)

    buf = StringIO.StringIO()
    with tarfile.open(None, 'w:gz', buf) as tf:
        tf.add(base, '.')

    os.environ['CI_CODE_SOURCE'] = 'tarball'
    os.environ['CI_PAYLOAD_URL'] = ds.put_file('payload.tgz', buf.getvalue())
    # the django charm uses branch for both meanings of payload
    os.environ['CI_BRANCH'] = os.environ['CI_PAYLOAD_URL']


def _get_args(args=None):
    parser = argparse.ArgumentParser(
        description='use juju-deployer to deploy services')
    parser.add_argument('--branch-payload', action='store_true',
                        help='''Use the current tip of this payload branch for
                        the code-base rather than a tarball of the current
                        directory.''')
    parser.add_argument('--reuse-payload', action='store_true',
                        help='''Use the already uploaded payload for
                        the code-base rather than a tarball of the current
                        directory. Use at your own risk.''')
    parser.add_argument('specific', metavar='<config.yaml>', nargs='*',
                        help='Use a specific list of juju-deployer configs.')

    return parser.parse_args(args)


def main():
    args = _get_args()

    check_environment()
    check_bootstrapped()
    check_dependencies()

    temp_dir = install_deployer()
    atexit.register(lambda: remove_branch(temp_dir))

    # A place to put our generated .yaml files.
    temp_dir = tempfile.mkdtemp()
    atexit.register(lambda: shutil.rmtree(temp_dir))
    working_dir = os.path.join(temp_dir, 'deployer')
    shutil.copytree(deployer_dir, working_dir)

    if not args.branch_payload and not args.reuse_payload:
        setup_swift()

    # Create .yaml files from .yaml.tmpl files.
    populate_templates(deployer_dir, working_dir)

    # Generate an argument list of -c FILE for each .yaml file in our working
    # directory.
    i = generate_argument_list(working_dir, args.specific)
    generated = list(chain.from_iterable(i))

    build_charms()

    args = ['juju-deployer', '-v'] + generated + ['ci-airline-staging']
    print 'Running {}'.format(' '.join(args))
    try:
        subprocess.check_call(args,
                              cwd=os.path.join(deployer_dir, '../charms'))
    except Exception as exc:
        print 'Problem deploying "ci-airline-staging": {}'.format(str(exc))

if __name__ == '__main__':
    dump_stack.install_stack_dump_signal()
    main()
