# Ubuntu CI Engine
# Copyright 2012-2013 Canonical Ltd.

# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License version 3, as
# published by the Free Software Foundation.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import urllib2

from django.conf import settings
from tastypie.api import Api
from tastypie.exceptions import ImmediateHttpResponse
from tastypie.http import (
    HttpAccepted,
    HttpApplicationError,
    HttpBadRequest,
    HttpNoContent,
    HttpNotFound
)
from tastypie.resources import ALL

from ppa_assigner import ppa_sync
from ppa_assigner.models import PPA

from ci_utils.tastypie.resource import AuthorizedResource


class PPAResource(AuthorizedResource):
    class Meta(AuthorizedResource.Meta):
        queryset = PPA.objects.filter()
        resource_name = 'ppa'

        filtering = {
            'state': ALL,
            'ticket_id': ALL,
            'timestamp': ALL,
        }

    def obj_create(self, bundle, **kwargs):
        '''We don't really create an object, so override tastypie.'''
        ticket_id = bundle.data.get('ticket_id')
        try:
            bundle.obj = PPA.reserve(ticket_id)
        except PPA.DoesNotExist as e:
            raise ImmediateHttpResponse(HttpNotFound(e.message))
        return bundle

    def patch_list(self, request, **kwargs):
        if request.body == '{"populate": true}':
            try:
                # populate is a bit different than normal "patch" operations
                PPA.populate_from_launchpad()
                return HttpAccepted()
            except urllib2.URLError as e:
                return HttpApplicationError(e)
        return super(PPAResource, self).patch_list(request, **kwargs)

    def hydrate_state(self, bundle):
        state = bundle.data['state']
        # tastypie by default will only let us refer to the state by their
        # numeric value. This is awkward for users, so this lets a user
        # specify something like "dirty" rather than 20.
        if type(state) != int:
            for val, label in PPA.STATES:
                if state == label:
                    bundle.data['state'] = val
                    break
        return bundle


class PPACopyResource(AuthorizedResource):
    class Meta(AuthorizedResource.Meta):
        resource_name = 'ppa_copy'
        # just return nothing to work with tastypie
        queryset = PPA.objects.none()

    def post_list(self, request, **kwargs):
        fmt = request.META.get('CONTENT_TYPE', 'application/json')
        data = self.deserialize(request, request.body, format=fmt)
        required = set(('source', 'target', 'series', 'packages'))
        missing = required - set(data.keys())
        if len(missing):
            return HttpBadRequest(
                'missing required parameters: %s\n' % ' '.join(required))

        ppa_sync.copy(
            data['source'], data['target'], data['packages'], data['series'])
        return HttpNoContent()


class PPAStatus(AuthorizedResource):
    class Meta(AuthorizedResource.Meta):
        resource_name = 'status'
        allowed_methods = ['get']
        # just return nothing to work with tastypie
        queryset = PPA.objects.none()

    def get_list(self, request, **kwargs):
        lp = settings.LAUNCHPAD_PPA_USER is not None
        lp = lp and settings.OAUTH_TOKEN is not None
        lp = lp and settings.OAUTH_TOKEN_SECRET is not None
        lp = lp and settings.OAUTH_CONSUMER_KEY is not None
        status = {
            'launchpad_configured': lp,
            'total ppas': PPA.objects.all().count(),
            'available ppas': PPA.objects.filter(state=PPA.AVAILABLE).count(),
        }
        return self.create_response(request, status)

v1_api = Api(api_name='v1')
v1_api.register(PPAResource())
v1_api.register(PPACopyResource())
v1_api.register(PPAStatus())
